"""
    Base Text2Vec Model
"""
import warnings
from ...base import Base2Vec
from abc import ABC, abstractmethod
from typing import Union, List, Dict

class BaseText2Vec(Base2Vec, ABC):
    def read(self, text: str):
        """An abstract method to specify the read method to read the data.
        """
        pass
    
    @property
    def test_word(self):
        return "dummy word"

    @property
    def vector_length(self):
        """
            Set the vector length of the model.
        """
        if hasattr(self, "_vector_length"):
            return getattr(self, "_vector_length") 
        else:
            print(f"The vector length is not explicitly stated so we are inferring " + \
                "from our test word - {self.test_word}.")
            setattr(self, "_vector_length", len(self.encode(self.test_word)))
            return self._vector_length
    
    @vector_length.setter
    def vector_length(self, value):
        self._vector_length = value
    
    @abstractmethod
    def encode(self, words: Union[List[str]]):
        pass


"""
Class for AutoEncoders.
"""
import warnings
from collections import defaultdict
from .models_dict import MODEL_REQUIREMENTS
import json

with warnings.catch_warnings():
    warnings.filterwarnings("ignore")
    from .encoders.text.tfhub import *
    from .encoders.text.vectorai import *
    from .encoders.text.tf_transformers import *
    from .encoders.text.torch_transformers import *
    from .encoders.text.sentence_transformers import *
    from .encoders.audio.vectorai import *
    from .encoders.audio.tfhub import *
    from .encoders.audio.pytorch import *
    from .encoders.image.tfhub import *
    from .encoders.image.vectorai import *
    from .encoders.image.fastai import *
    from .bi_encoders.text_text.tfhub import *
    from .bi_encoders.text_text.torch_transformers import *
    from .bi_encoders.text_text.sentence_transformers import *

# Include the class and then the requirements key from models_dict.py
ENCODER_MAPPINGS = defaultdict(tuple, {
    Albert2Vec.definition.model_id : (Albert2Vec, "encoders-text-tfhub-albert"),
    Bert2Vec.definition.model_id : (Bert2Vec, "encoders-text-tfhub-bert"),
    Elmo2Vec.definition.model_id: (Elmo2Vec, "encoders-text-tfhub-elmo"),
    LaBSE2Vec.definition.model_id : (LaBSE2Vec, "encoders-text-tfhub-labse"),
    LegalBert2Vec.definition.model_id: (LegalBert2Vec, "encoders-text-torch-transformers-legalbert"),
    Longformer2Vec.definition.model_id: (Longformer2Vec, "encoders-text-torch-transformers-longformer"),
    USE2Vec.definition.model_id : (USE2Vec, "encoders-text-tfhub-use"),
    USEMulti2Vec.definition.model_id : (USEMulti2Vec, "encoders-text-tfhub-use"),
    USELite2Vec.definition.model_id : (USELite2Vec, "encoders-text-tfhub-use"),
    

    # "text/tf-transformers" : (TFTransformer2Vec, "encoders-text-tf-transformers"),
    # "text/torch-transformers" : (Transformer2Vec, "encoders-text-torch-transformers"),

    Wav2Vec.definition.model_id : (Wav2Vec, "encoders-audio-pytorch-fairseq"),
    SpeechEmbedding2Vec.definition.model_id : (SpeechEmbedding2Vec, "encoders-audio-tfhub-speech_embedding"),
    Trill2Vec.definition.model_id : (Trill2Vec, 'encoders-audio-tfhub-trill'),
    TrillDistilled2Vec.definition.model_id : (TrillDistilled2Vec, 'encoders-audio-tfhub-trill'),
    Vggish2Vec.definition.model_id : (Vggish2Vec, 'encoders-audio-tfhub-vggish'),
    Yamnet2Vec.definition.model_id : (Yamnet2Vec, "encoders-audio-tfhub-vggish"),

    BitSmall2Vec.definition.model_id : (BitSmall2Vec, "encoders-image-tfhub-bit"),
    BitMedium2Vec.definition.model_id : (BitMedium2Vec, "encoders-image-tfhub-bit"),
    InceptionV12Vec.definition.model_id : (InceptionV12Vec, "encoders-image-tfhub-inception"),
    InceptionV22Vec.definition.model_id : (InceptionV22Vec, "encoders-image-tfhub-inception"),
    InceptionV32Vec.definition.model_id : (InceptionV32Vec, "encoders-image-tfhub-inception"),
    InceptionResnet2Vec.definition.model_id : (InceptionResnet2Vec, "encoders-image-tfhub-inception-resnet"),
    MobileNetV12Vec.definition.model_id : (MobileNetV12Vec, "encoders-image-tfhub-mobilenet"),
    MobileNetV22Vec.definition.model_id : (MobileNetV22Vec, "encoders-image-tfhub-mobilenet"),
    ResnetV12Vec.definition.model_id : (ResnetV12Vec, "encoders-image-tfhub-resnet"),
    ResnetV22Vec.definition.model_id : (ResnetV22Vec, "encoders-image-tfhub-resnet"),
    FastAIResnet2Vec.definition.model_id: (FastAIResnet2Vec, "encoders-image-fastai-resnet"),
})

class AutoEncoder:
    """
        The AutoEncoder class. To view the AutoEncoder's model, this means 

        Example: 
            >>> from vectorhub.auto_encoder import *
            >>> model = AutoEncoder.from_model('text/bert')
            >>> # To view a list of models use the follow 
            >>> list_all_auto_models()
    """
    def __init__(self):
        pass

    @staticmethod
    def from_model(model_id, *args, **kwargs):
        """
            The AutoEncoder class. To view the AutoEncoder's model, this means 
            Args:
                model_id: The ID of the model - these can be found in list_all_auto_models()
                args: The arguments for the model instantiation
                kwargs: The keyword arguments for the model instantiation
            Example: 
                >>> from vectorhub.auto_encoder import *
                >>> model = AutoEncoder.from_model('text/bert')
                >>> # To view a list of models use the follow 
                >>> list_all_auto_models()
        """
        model_callable, requirements = ENCODER_MAPPINGS[model_id]
        assert is_all_dependency_installed(MODEL_REQUIREMENTS[requirements]), "Missing requirements! Please install."
        model = model_callable(*args, **kwargs)
        return model

BIENCODER_MAPPINGS = {
    USEMultiQA2Vec.definition.model_id : (USEMultiQA2Vec, "text-bi-encoder-tfhub-use-qa"),
    USEQA2Vec.definition.model_id : (USEQA2Vec, "text-bi-encoder-tfhub-use-qa"),
    LAReQA2Vec.definition.model_id: (LAReQA2Vec, "text-bi-encoder-tfhub-lareqa-qa"),
    DPR2Vec.definition.model_id : (DPR2Vec, "text-bi-encoder-torch-dpr")
}

class AutoBiEncoder:
    def __init__(self):
        pass
    

    @staticmethod
    def from_model(model_id, *args, **kwargs):
        model_callable, requirements = BIENCODER_MAPPINGS[model_id]
        assert is_all_dependency_installed(MODEL_REQUIREMENTS[requirements]), "Missing requirements! Please install."
        model = model_callable(*args, **kwargs)
        return model


def list_all_auto_models():
    """
        List all available models available for auto models.
    """
    return list(ENCODER_MAPPINGS.keys()) + list(BIENCODER_MAPPINGS.keys())

def get_model_definitions(json_fn='models.json'):
    """
        Get a list of dictionaries with the definitions of the models.
    """
    all_models = list()
    global_vars = globals().items()
    for k, v in global_vars:
        if hasattr(v, 'definition'):
            values = v.definition.to_dict()
            values.update({"_id": values['model_id']})
            all_models.append(values)
    if json_fn is not None:
        with open(json_fn, 'w') as f:
            json.dump(all_models, f)
    else:
        return all_models



"""
    Dictionary For Models
"""

# Map Model to the requirements here.
# This is used to allow the user to list what models they have which can be installed.

MODEL_REQUIREMENTS = {
    "encoders-audio-pytorch-fairseq": "encoders-audio-pytorch",
    "encoders-audio-tfhub-speech_embedding": "encoders-audio-tfhub",
    "encoders-audio-tfhub-trill": 'encoders-audio-tfhub',
    "encoders-audio-tfhub-vggish": "encoders-audio-tfhub",
    "encoders-audio-tfhub-yamnet": "encoders-audio-tfhub", 
    "encoders-audio-vectorai": None,

    "encoders-text-tfhub-albert": "encoders-text-tfhub",
    "encoders-text-tfhub-bert": "encoders-text-tfhub-windows",
    "encoders-text-tfhub-elmo": "encoders-text-tfhub",
    "encoders-text-tfhub-labse": "encoders-text-tfhub-windows",
    "encoders-text-tfhub-use": "encoders-text-tfhub-windows", 
    "encoders-text-tfhub-use-multi": "encoders-text-tfhub", 
    "encoders-text-torch-transformers-legalbert": "encoders-text-torch-transformers",
    "encoders-text-tf-transformers-auto": "encoders-text-tf-transformers",
    "encoders-text-torch-transformers-auto": "encoders-text-torch-transformers",
    "encoders-text-torch-transformers-longformer": "encoders-text-torch-transformers",
    "encoders-text-sentence-transformers": "encoders-text-sentence-transformers",
    "encoders-text-vectorai": None,
    
    
    "encoders-image-tfhub-bit": "encoders-image-tfhub",
    "encoders-image-tfhub-resnet": "encoders-image-tfhub",
    "encoders-image-tfhub-inception": "encoders-image-tfhub",
    "encoders-image-tfhub-inception-resnet": "encoders-image-tfhub",
    "encoders-image-tfhub-mobilenet": "encoders-image-tfhub",
    "encoders-image-fastai-resnet": "encoders-image-fastai",
    "encoders-image-vectorai": None,
    
    "text-bi-encoder-tfhub-lareqa-qa": "encoders-text-tfhub",
    "text-bi-encoder-tfhub-use-qa": "encoders-text-tfhub",
    "text-bi-encoder-torch-dpr": "encoders-text-torch-transformers",
    
}



from math import ceil
import numpy as np
import os
import tempfile
from ...import_utils import *

if is_all_dependency_installed('encoders-video'):
    import librosa
    import soundfile as sf
    from cv2 import cv2
    from moviepy.video.io.ffmpeg_reader import ffmpeg_parse_infos
    from moviepy.video.io.VideoFileClip import VideoFileClip


class FrameSamplingFilter():

    def __init__(self, every=None, hertz=None, top_n=None):
        if every is None and hertz is None and top_n is None:
            raise ValueError("When initializing the FrameSamplingFilter, "
                             "one of the 'every', 'hertz', or 'top_n' must "
                             "be specified.")
        self.every = every
        self.hertz = hertz
        self.top_n = top_n

    def get_audio_sampling_rate(self, filename: str):
        infos = ffmpeg_parse_infos(filename)
        fps = infos.get('audio_fps', 44100)
        if fps == 'unknown':
            fps = 44100
        return fps

    def load_clip(self, filename: str):
        audio_fps = self.get_audio_sampling_rate(filename)
        self.clip = VideoFileClip(filename, audio_fps)

    def initialize_video(self, filename: str):
        self.filename = filename
        self.load_clip(filename)
        self.fps = self.clip.fps
        self.width = self.clip.w
        self.height = self.clip.h
        self.frame_index = range(int(ceil(self.fps * self.clip.duration)))
        self.duration = self.clip.duration
        self.n_frames = len(self.frame_index)

    def get_audio_vector(self, new_sampling_rate: int = 16000):
        fd, fp = tempfile.mkstemp()
        audio = f'{fp}.wav'
        self.clip.audio.to_audiofile(audio)
        data, sampling_rate = sf.read(audio, dtype='float32')
        os.close(fd)
        os.remove(audio)
        return np.array(librosa.resample(data.T, sampling_rate, new_sampling_rate))

    def transform(self, filename: str):
        self.initialize_video(filename)

        if (self.every is not None):
            new_idx = range(self.n_frames)[::self.every]
        elif (self.hertz is not None):
            interval = self.fps / float(self.hertz)
            new_idx = np.arange(0, self.n_frames, interval).astype(int)
            new_idx = list(new_idx)
        elif self.top_n is not None:
            diffs = []
            for i, img in enumerate(range(self.n_frames)):
                if i == 0:
                    last = img
                    continue
                pixel_diffs = cv2.sumElems(cv2.absdiff(
                    self.get_frame(last), self.get_frame(img)))
                diffs.append(sum(pixel_diffs))
                last = img
            new_idx = sorted(range(len(diffs)),
                             key=lambda i: diffs[i],
                             reverse=True)[:self.top_n]

        result = []
        for index in new_idx:
            result.append(self.get_frame(index))
        return result

    def get_frame(self, index: int):
        return self.clip.get_frame(index)

    def iter_frames(self):
        for i, f in enumerate(self.frame_index):
            yield self.get_frame(f)



import warnings
import numpy as np
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import is_all_dependency_installed
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseText2Vec
from .use import USE2Vec

if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tfhub-use-multi']):
    import tensorflow as tf
    from tensorflow.python.framework.errors_impl import NotFoundError
    try:
        import tensorflow_text
    except NotFoundError:
        print('The installed Tensorflow Text version is not aligned with tensorflow, make sure that tensorflow-text version is same version as tensorflow')

USELiteModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tfhub/use_lite.md')

class USELite2Vec(BaseText2Vec):
    definition = USELiteModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/universal-sentence-encoder-lite/2'):
        list_of_urls = [
            "https://tfhub.dev/google/universal-sentence-encoder-lite/2",
        ]
        self.validate_model_url(model_url, list_of_urls)
        self.vector_length = 512
        warnings.warn("Using USELite2Vec requires disabling tf2 behaviours: tf.disable_v2_behavior(). Meaning it can break the usage of other models if ran. If you are ok with this run model.init() to disable tf2 and run USELite2Vec")

    def init(self):
        import tensorflow.compat.v1 as tf
        tf.disable_v2_behavior()
        self.input_placeholder = tf.sparse_placeholder(tf.int64, shape=[None, None])
        self.model = hub.Module("https://tfhub.dev/google/universal-sentence-encoder-lite/2")
        self.encodings = module(inputs=dict(
            values=self.input_placeholder.values,
            indices=self.input_placeholder.indices,
            dense_shape=self.input_placeholder.dense_shape
        ))
        with tf.Session() as sess:
            spm_path = sess.run(module(signature="spm_path"))
        self.sp = spm.SentencePieceProcessor()
        self.sp.Load(spm_path)

    def process_texts(self, texts):
        ids = [self.sp.EncodeAsIds(x) for x in texts]
        return (
            [item for sublist in ids for item in sublist], 
            [[row,col] for row in range(len(ids)) for col in range(len(ids[row]))], 
            (len(ids), max(len(x) for x in ids))
        )

    @catch_vector_errors
    def encode(self, text):
        values, indices, dense_shape = self.process_texts([text])
        with tf.Session() as session:
            session.run([tf.global_variables_initializer(), tf.tables_initializer()])
            message_embeddings = session.run(self.encodings,
                feed_dict={self.input_placeholder.values: values,
                            self.input_placeholder.indices: indices,
                            self.input_placeholder.dense_shape: dense_shape})
        return np.array(message_embeddings)[0].tolist()

    @catch_vector_errors
    def bulk_encode(self, texts, threads=10, chunks=100):
        values, indices, dense_shape = self.process_texts(texts)
        with tf.Session() as session:
            session.run([tf.global_variables_initializer(), tf.tables_initializer()])
            message_embeddings = session.run(self.encodings,
                feed_dict={self.input_placeholder.values: values,
                            self.input_placeholder.indices: indices,
                            self.input_placeholder.dense_shape: dense_shape})
        return np.array(message_embeddings).tolist()


import functools
import warnings
import traceback
import numpy as np
from .errors import ModelError
from typing import Any, List
from abc import ABC, abstractmethod

BASE_2VEC_DEFINITON = {
    "vector_length": None,
    "description": None,
    "paper": None,
    "repo": None,
    "model_name": None,
    "architecture": None,
    "tasks": None,
    "limitations": None,
    "download_required": None,
    "training_required": None,
    "finetunable": None,
}

def catch_vector_errors(func):
    """
        Decorate function and avoid vector errors.
        Example:
            class A:
                @catch_vector_errors
                def encode(self):
                    return [1, 2, 3]
    """
    @functools.wraps(func)
    def catch_vector(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except:
            warnings.warn("Unable to encode. Filling in with dummy vector.")
            traceback.print_exc()
            # get the vector length from the self body
            vector_length = args[0].vector_length
            if isinstance(args[1], str):
                return [1e-7] * vector_length
            elif isinstance(args[1], list):
                # Return the list of vectors
                return [[1e-7] * vector_length] * len(args[1])
            else:
                return [1e-7] * vector_length
    return catch_vector

class Base2Vec:
    """
        Base class for vector
    """
    def __init__(self):
        self.__dict__.update(BASE_2VEC_DEFINITON)

    @classmethod
    def validate_model_url(cls, model_url: str, list_of_urls: List[str]):
        """
            Validate the model url belongs in the list of urls. This is to help 
            users to avoid mis-spelling the name of the model.

            # TODO:
            Improve model URL validation to not include final number in URl string.
            Args:
                model_url: The URl of the the model in question
                list_of_urls: The list of URLS for the model in question
        
        """
        if(model_url in list_of_urls):
            return True
        raise ModelError(
            message="We currently not support this url. If issue persist then contact us.")
            
    @classmethod
    def chunk(self, lst: List, chunk_size: int):
        """
        Chunk an iterable object in Python but not a pandas DataFrame.
        Args:
            lst:
                Python List
            chunk_size:
                The chunk size of an object.
        Example:
            >>> documents = [{...}]
            >>> ViClient.chunk(documents)
        """
        for i in range(0, len(lst), chunk_size):
            yield lst[i: i + chunk_size]

    def _vector_operation(self, vectors, vector_operation: str = "mean", axis=0):
        """
            Args:
                Vectors: the list of vectors to include 
                vector_operation: One of ['mean', 'minus', 'sum', 'min', 'max']
                axis: The axis to which to perform the operation
        """
        if vector_operation == "mean":
            return np.mean(vectors, axis=axis).tolist()
        elif vector_operation == 'minus':
            return np.subtract(vectors, axis=axis).tolist()
        elif vector_operation == "sum":
            return np.sum(vectors, axis=axis).tolist()
        elif vector_operation == "min":
            return np.min(vectors, axis=axis).tolist()
        elif vector_operation == "max":
            return np.max(vectors, axis=axis).tolist()
        else:
            return np.mean(vectors, axis=axis).tolist()

    @property
    def __name__(self):
        """
            Return the name of the model. If name is not set, returns the
            model_id.
        """
        if hasattr(self, '_name'):
            return self._name
        elif hasattr(self, 'definition'):
            if '/' in self.definition.model_id:
                return self.definition.model_id.split('/')[1]
            return self.definition.model_id
        return ''
    
    @__name__.setter        
    def __name__(self, value):
        """
            Set the name.
        """
        setattr(self, '_name', value)



from ...encoders.text.base import BaseText2Vec
from abc import ABC, abstractmethod

class BaseTextText2Vec(BaseText2Vec, ABC):
    def encode(self):
        pass

    @abstractmethod
    def encode_question(self):
        pass

    @abstractmethod
    def encode_answer(self):
        pass



"""
    The base class for FastAI as much of it can be replaced easily by changing the model.
"""
from abc import abstractproperty
from ..base import BaseImage2Vec
from ....import_utils import is_all_dependency_installed
from ....base import catch_vector_errors

if is_all_dependency_installed('encoders-image-fastai'):
    import torch
    import numpy as np
    # We use wildcard imports for FastAI as this is the way it is handled in the documentation.
    from fastai.vision.all import cnn_learner, resnet34, ImageDataLoaders, error_rate, hook_outputs
    from fastai.data.all import get_image_files, untar_data, URLs

class FastAIBase(BaseImage2Vec):
    def __init__(self, databunch=None, architecture=None):
        self.databunch = databunch
        self.architecture = architecture
        self._create_learner(databunch, architecture)
    
    def _instantiate_empty_dataloader(self):
        """
            As it is almost impossible to instantiate an empty dataloader, we use a CIFAR as a dummy.
        """
        path = untar_data(URLs.CIFAR_100)
        files = get_image_files(path/"train")
        return ImageDataLoaders.from_lists(path, files, [0] * len(files))
    
    def _create_learner(self):
        dls = self._instantiate_empty_dataloader()
        self.learn = cnn_learner(dls, resnet34, metrics=error_rate)

    @abstractproperty
    def extraction_layer(self):
        pass

    @catch_vector_errors
    def encode(self, image):
        with hook_outputs(self.extraction_layer) as h:
            if isinstance(image, str):
                y = self.learn.predict(self.read(image))
            elif isinstance(image, (np.ndarray, np.generic)):
                y = self.learn.predict(image)
        return h.stored[0].cpu().numpy().tolist()[0]
    
    @catch_vector_errors
    def bulk_encode(self, images):
        return [self.encode(x) for x in images]



"""Errors
"""

class ModelError(Exception):
    """Base error class for all errors in library
    """

    def __init__(self, message: str):
        """
        The main Vector Hub base error.
        Args:
            message: The error message
        Example:
            >>> raise ModelError("Missing ____.")
        """
        self.response_message = message


"""
    Abstract from: https://arxiv.org/abs/1409.4842

    We propose a deep convolutional neural network architecture codenamed "Inception", which was responsible for setting the new state of the art for classification and detection in the ImageNet Large-Scale Visual Recognition Challenge 2014 (ILSVRC 2014). The main hallmark of this architecture is the improved utilization of the computing resources inside the network. This was achieved by a carefully crafted design that allows for increasing the depth and width of the network while keeping the computational budget constant. To optimize quality, the architectural decisions were based on the Hebbian principle and the intuition of multi-scale processing. One particular incarnation used in our submission for ILSVRC 2014 is called GoogLeNet, a 22 layers deep network, the quality of which is assessed in the context of classification and detection.

"""
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ..base import BaseImage2Vec
from .inceptionv1 import InceptionV12Vec

if is_all_dependency_installed('encoders-image-tfhub'):
    import tensorflow as tf
    import tensorflow_hub as hub

InceptionV2ModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/inceptionv2')

__doc__ = InceptionV2ModelDefinition.create_docs()

class InceptionV22Vec(InceptionV12Vec):
    definition = InceptionV2ModelDefinition
    def __init__(self, model_url: str = 'https://tfhub.dev/google/imagenet/inception_v2/feature_vector/4'):
        self.model_url = model_url
        self.init(model_url)
        self.vector_length = 1024

    @property
    def urls(self):
        """
        URLS for inception.
        """
        return {
            'https://tfhub.dev/google/imagenet/inception_v2/feature_vector/4': {'vector_length': 1024}
        }



from typing import List
from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
from ..base import BaseTextText2Vec
from .use_qa import USEQA2Vec
if is_all_dependency_installed(MODEL_REQUIREMENTS['text-bi-encoder-tfhub-use-qa']):
    import bert
    import numpy as np
    import tensorflow as tf
    import tensorflow_hub as hub
    from tensorflow.python.framework.errors_impl import NotFoundError
    try:
        import tensorflow_text
    except NotFoundError:
        print('The installed Tensorflow Text version is not aligned with tensorflow, make sure that tensorflow-text version is same version as tensorflow')

USEMultiQAModelDefinition = ModelDefinition(markdown_filepath='bi_encoders/text_text/tfhub/use_multi_qa')

class USEMultiQA2Vec(USEQA2Vec):
    definition = USEMultiQAModelDefinition
    def __init__(self):
        self.model_url = "https://tfhub.dev/google/universal-sentence-encoder-multilingual-qa/3"
        self.model = hub.load(self.model_url)
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.vector_length = 512



from typing import List
from ..base import BaseText2Vec
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ....models_dict import MODEL_REQUIREMENTS
if is_all_dependency_installed(MODEL_REQUIREMENTS['encoders-text-tf-transformers-auto']):
    import tensorflow as tf
    from transformers import AutoTokenizer, TFAutoModel

TransformerModelDefinition = ModelDefinition(markdown_filepath='encoders/text/tf_transformers/tf_auto_transformers.md')

__doc__ = TransformerModelDefinition.create_docs()


class TFTransformer2Vec(BaseText2Vec):
    definition = TransformerModelDefinition
    def __init__(self, model_name: str, config=None):
        if config is None:
            self.model = TFAutoModel.from_pretrained(model_name)
        else:
            self.model = TFAutoModel.from_pretrained(model_name, config=config)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
    
    @catch_vector_errors
    def encode(self, text: str) -> List[float]:
        """
            Encode word from transformers.
            This takes the beginning set of tokens and turns them into vectors
            and returns mean pooling of the tokens.
            Args:
                word: string 
        """
        return tf.reduce_mean(self.model(self.tokenizer(text, return_tensors='tf'))[0], axis=1).numpy().tolist()[0]
    
    @catch_vector_errors
    def bulk_encode(self, texts: List[str]) -> List[List[float]]:
        """
            Bulk encode words from transformers.
        """
        return tf.reduce_mean(self.model(self.tokenizer(texts, return_tensors='tf', truncation=True, padding=True))[0], axis=1).numpy().tolist()



from datetime import date
from ....base import catch_vector_errors
from ....doc_utils import ModelDefinition
from ....import_utils import *
from ..base import BaseImage2Vec

if is_all_dependency_installed('encoders-image-tfhub'):
    import tensorflow as tf
    import tensorflow_hub as hub

BITModelDefinition = ModelDefinition(markdown_filepath='encoders/image/tfhub/bit')

__doc__ = BITModelDefinition.create_docs()

class BitSmall2Vec(BaseImage2Vec):
    definition = BITModelDefinition
    def __init__(self, model_url: str = "https://tfhub.dev/google/bit/s-r50x1/1"):
        self.validate_model_url(model_url, self.list_of_urls)
        self.init(model_url)
        self.vector_length = self.list_of_urls[model_url]["vector_length"]

    @property
    def list_of_urls(self):
        return {
            'https://tfhub.dev/google/bit/s-r50x1/1': {"vector_length":2048}, # 2048 output shape
            'https://tfhub.dev/google/bit/s-r50x3/1': {"vector_length":6144},   # 6144 output shape
            'https://tfhub.dev/google/bit/s-r101x1/1': {"vector_length":2048},  # 2048 output shape
            'https://tfhub.dev/google/bit/s-r101x3/1': {"vector_length":6144},  # 6144 output shape
            'https://tfhub.dev/google/bit/s-r152x4/1': {"vector_length":8192},   # 8192 output shape
        }

    @property
    def urls(self):
        """
        Get the urls and their vector length.
        """
        return {
            'https://tfhub.dev/google/bit/s-r50x1/1': {"vector_length":2048}, # 2048 output shape
            'https://tfhub.dev/google/bit/s-r50x3/1': {"vector_length":6144},   # 6144 output shape
            'https://tfhub.dev/google/bit/s-r101x1/1': {"vector_length":2048},  # 2048 output shape
            'https://tfhub.dev/google/bit/s-r101x3/1': {"vector_length":6144},  # 6144 output shape
            'https://tfhub.dev/google/bit/s-r152x4/1': {"vector_length":8192},   # 8192 output shape
        }

    def init(self, model_url: str):
        self.model_url = model_url
        self.model_name = self.model_url.replace(
            'https://tfhub.dev/google/', '').replace('/', '_')
        self.model = hub.load(self.model_url)

    @catch_vector_errors
    def encode(self, image):
        if isinstance(image, str):
            image = self.read(image)
        return self.model([image]).numpy().tolist()[0]

    @catch_vector_errors
    def bulk_encode(self, images):
        """
            Bulk encode. Chunk size should be specified outside of the images.
        """
        # TODO: Change from list comprehension to properly read
        return [self.encode(x) for x in images]



from typing import Union
from ...base import Base2Vec
from ...import_utils import is_all_dependency_installed
# TODO: Change encoders-image-tfhub into general encoders-image
if is_all_dependency_installed('encoders-image'):
    import io
    import imageio
    import numpy as np
    import matplotlib.pyplot as plt
    from urllib.request import urlopen, Request
    from urllib.parse import quote
    from skimage import transform

class BaseImage2Vec(Base2Vec):
    def read(self, image: str):
        """
            An method to read images. 
            Args:
                image: An image link/bytes/io Bytesio data format.
                as_gray: read in the image as black and white
        """
        if type(image) == str:
            if 'http' in image:
                b = io.BytesIO(urlopen(Request(
                    quote(image, safe=':/?*=\''), headers={'User-Agent': "Mozilla/5.0"})).read())
            else:
                b = image
        elif type(image) == bytes:
            b = io.BytesIO(image)
        elif type(image) == io.BytesIO:
            b = image
        else:
            raise ValueError("Cannot process data type. Ensure it is is string/bytes or BytesIO.")
        try:
            return np.array(imageio.imread(b, pilmode="RGB"))
        # TODO: Flesh out exceptions
        except:
            return np.array(imageio.imread(b)[:, :, :3])
    
    def to_grayscale(self, sample, rgb_weights: list=None):
        """
            Converting an image from RGB to Grayscale
        """
        if rgb_weights is None:
            return np.repeat(np.dot(sample[...,:3], self.rgb_weights)[..., np.newaxis], 3, -1)
        else:
            return np.repeat(np.dot(sample[...,:3], rgb_weights)[..., np.newaxis], 3, -1)
    
    @property
    def rgb_weights(self):
        """
            Get RGB weights for grayscaling.
        """
        return [0.2989, 0.5870, 0.1140]

    def show_image(self, sample, cmap=None, is_grayscale=True):
        """
            Show an image once it is read. 
            Arg:
                sample: Image that is read (numpy array)
        """
        if is_grayscale:
            return plt.imshow(sample, cmap=plt.get_cmap("gray"))
        return plt.imshow(sample, cmap=cmap)
    
    def image_resize(self, image_array, width=0, height=0, rescale=0, resize_mode='symmetric'):
        if width and height:
            image_array = transform.resize(image_array, (width, height), mode=resize_mode, preserve_range=True)
        if rescale:
            image_array = transform.rescale(image_array, rescale, preserve_range=True, anti_aliasing=True)
        return np.array(image_array)
